use std::{num::NonZeroU64, path::PathBuf};

use cb::TriangleRenderResources;
use eframe::{
    egui, egui_wgpu,
    wgpu::{self, util::DeviceExt},
};
use egui_file_dialog::FileDialog;
use nalgebra::{Matrix3, SMatrix};

use crate::cb::CustomTriangleCallback;

pub struct App {
    color_matrix: SMatrix<f32, 4, 4>,
    pos_matrix: SMatrix<f32, 3, 3>,
    file_dialog: FileDialog,
    new_file: Option<PathBuf>,
    cur_img_size: Option<egui::Vec2>,
    image_loaded: bool,
}
impl App {
    pub fn new<'a>(cc: &'a eframe::CreationContext<'a>) -> Option<Self> {
        // Get the WGPU render state from the eframe creation context. This can also be retrieved
        // from `eframe::Frame` when you don't have a `CreationContext` available.
        let wgpu_render_state = cc.wgpu_render_state.as_ref()?;

        let device = &wgpu_render_state.device;

        // let cat = include_bytes!("./cat.jpg");
        // let cat_img = image::load_from_memory(cat).unwrap();
        // let cat_rgba = cat_img.to_rgba();

        // use image::GenericImageView;
        // let dimensions = cat_img.dimensions();

        // let tex_size = wgpu::Extent3d {
        //     width: dimensions.0,
        //     height: dimensions.1,
        //     depth_or_array_layers: 1,
        // };
        // let cat_tex = device.create_texture(&wgpu::TextureDescriptor {
        //     size: tex_size,
        //     mip_level_count: 1,
        //     sample_count: 1,
        //     dimension: wgpu::TextureDimension::D2,
        //     format: wgpu::TextureFormat::Rgba8UnormSrgb,
        //     usage: wgpu::TextureUsages::TEXTURE_BINDING | wgpu::TextureUsages::COPY_DST,
        //     label: Some("cat"),
        //     view_formats: &[],
        // });

        let shader = device.create_shader_module(wgpu::ShaderModuleDescriptor {
            label: Some("shader"),
            source: wgpu::ShaderSource::Wgsl(include_str!("./shader.wgsl").into()),
        });

        let tex_bind_group_layout =
            device.create_bind_group_layout(&wgpu::BindGroupLayoutDescriptor {
                label: Some("tex_bind_group_layout"),
                entries: &[
                    wgpu::BindGroupLayoutEntry {
                        binding: 0,
                        visibility: wgpu::ShaderStages::FRAGMENT,
                        ty: wgpu::BindingType::Texture {
                            multisampled: false,
                            view_dimension: wgpu::TextureViewDimension::D2,
                            sample_type: wgpu::TextureSampleType::Float { filterable: true },
                        },
                        count: None,
                    },
                    wgpu::BindGroupLayoutEntry {
                        binding: 1,
                        visibility: wgpu::ShaderStages::FRAGMENT,
                        // This should match the filterable field of the
                        // corresponding Texture entry above.
                        ty: wgpu::BindingType::Sampler(wgpu::SamplerBindingType::Filtering),
                        count: None,
                    },
                ],
            });
        let matrices_bind_group_layout =
            device.create_bind_group_layout(&wgpu::BindGroupLayoutDescriptor {
                label: Some("matrices_bind_group_layout"),
                entries: &[
                    wgpu::BindGroupLayoutEntry {
                        binding: 0,
                        visibility: wgpu::ShaderStages::FRAGMENT,
                        ty: wgpu::BindingType::Buffer {
                            ty: wgpu::BufferBindingType::Uniform,
                            has_dynamic_offset: false,
                            min_binding_size: Some(NonZeroU64::new(64)?),
                        },
                        count: None,
                    },
                    wgpu::BindGroupLayoutEntry {
                        binding: 1,
                        visibility: wgpu::ShaderStages::FRAGMENT,
                        ty: wgpu::BindingType::Buffer {
                            ty: wgpu::BufferBindingType::Uniform,
                            has_dynamic_offset: false,
                            min_binding_size: Some(NonZeroU64::new(36)?),
                        },
                        count: None,
                    },
                ],
            });

        let pipeline_layout = device.create_pipeline_layout(&wgpu::PipelineLayoutDescriptor {
            label: Some("pipeline_layout"),
            bind_group_layouts: &[&tex_bind_group_layout, &matrices_bind_group_layout],
            push_constant_ranges: &[],
        });

        let pipeline = device.create_render_pipeline(&wgpu::RenderPipelineDescriptor {
            label: Some("custom3d"),
            layout: Some(&pipeline_layout),
            vertex: wgpu::VertexState {
                module: &shader,
                entry_point: "vs_main",
                buffers: &[],
                // compilation_options: wgpu::PipelineCompilationOptions::default(),
            },
            fragment: Some(wgpu::FragmentState {
                module: &shader,
                entry_point: "fs_main",
                targets: &[Some(wgpu_render_state.target_format.into())],
                // compilation_options: wgpu::PipelineCompilationOptions::default(),
            }),
            primitive: wgpu::PrimitiveState::default(),
            depth_stencil: None,
            multisample: wgpu::MultisampleState::default(),
            multiview: None,
        });

        // let color_matrix_buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
        //     label: Some("color_matrix_buffer"),
        //     contents: bytemuck::cast_slice(&[0.0_f32; 4]), // 16 bytes aligned!
        //     // Mapping at creation (as done by the create_buffer_init utility) doesn't require us to to add the MAP_WRITE usage
        //     // (this *happens* to workaround this bug )
        //     usage: wgpu::BufferUsages::COPY_DST | wgpu::BufferUsages::UNIFORM,
        // });

        // let bind_group = device.create_bind_group(&wgpu::BindGroupDescriptor {
        //     label: Some("custom3d"),
        //     layout: &bind_group_layout,
        //     entries: &[wgpu::BindGroupEntry {
        //         binding: 0,
        //         resource: uniform_buffer.as_entire_binding(),
        //     }],
        // });

        // Because the graphics pipeline must have the same lifetime as the egui render pass,
        // instead of storing the pipeline in our `Custom3D` struct, we insert it into the
        // `paint_callback_resources` type map, which is stored alongside the render pass.

        // wgpu_render_state
        //     .renderer
        //     .write()
        //     .callback_resources
        //     .insert(TriangleRenderResources {
        //         pipeline,
        //         bind_group,
        //         uniform_buffer,
        //     });

        Some(Self {
            color_matrix: SMatrix::identity(),
            pos_matrix: SMatrix::identity(),
            file_dialog: FileDialog::new(),
            new_file: None,
            image_loaded: false,
            cur_img_size: None,
        })
    }
}

impl eframe::App for App {
    fn update(&mut self, ctx: &eframe::egui::Context, frame: &mut eframe::Frame) {
        let mut img = None;
        egui::SidePanel::right("sidebar").show(ctx, |ui| {
            if ui.button("Select Image").clicked() {
                self.file_dialog.select_file()
            }
            if self.file_dialog.update(ctx).selected().is_some() {
                if let Some(path) = self.file_dialog.take_selected() {
                    println!("path: {path:?}");
                    if path.ends_with(".jpg") || path.ends_with(".jpeg") || path.ends_with(".png") {
                        img = Some(image::open(path).unwrap().to_rgba8());
                    }
                }
            }

            ui.add_space(8.);
            ui.label("Color Matrix Editor");
            components::mat_editor(ui, &mut self.color_matrix, "color_matrix");
            ui.add_space(8.);
            ui.label("Position Matrix Editor");
            components::mat_editor(ui, &mut self.pos_matrix, "pos_matrix");
        });
        egui::CentralPanel::default().show(ctx, |ui| {
            ui.heading("My egui Application");

            if let Some(dimensions) = self.cur_img_size {
                egui::Frame::canvas(ui.style()).show(ui, |ui| {
                    let (rect, response) = ui.allocate_exact_size(dimensions, egui::Sense::drag());

                    // let callback = egui_wgpu::Callback::new_paint_callback(
                    //     rect,
                    //     CustomTriangleCallback { angle: 0.5 },
                    // );
                    // ui.painter().add(callback);
                });
            }
        });
    }
}

mod gpu_proc {
    use eframe::wgpu;

    pub struct GpuResources {
        pipeline: wgpu::RenderPipeline,
        // img_buffer: Option<wgpu::
    }
}

mod components {
    use eframe::egui;
    use nalgebra::{Matrix, SMatrix};

    pub fn mat_editor<const R: usize, const C: usize>(
        ui: &mut egui::Ui,
        mat: &mut SMatrix<f32, R, C>,
        id: &str,
    ) -> egui::Response {
        egui::Grid::new(id)
            .show(ui, |ui| {
                mat.row_iter_mut().enumerate().for_each(|(i, mut row)| {
                    row.iter_mut().for_each(|item| {
                        ui.add(
                            egui::DragValue::new(item)
                                .speed(0.01)
                                .clamp_range(-1.0..=1.0),
                        );
                    });
                    ui.end_row();
                })
            })
            .response
    }
}

mod cb {
    use eframe::{egui, egui_wgpu, wgpu};

    pub struct CustomTriangleCallback {
        pub angle: f32,
    }

    impl egui_wgpu::CallbackTrait for CustomTriangleCallback {
        fn prepare(
            &self,
            device: &wgpu::Device,
            queue: &wgpu::Queue,
            _screen_descriptor: &egui_wgpu::ScreenDescriptor,
            _egui_encoder: &mut wgpu::CommandEncoder,
            resources: &mut egui_wgpu::CallbackResources,
        ) -> Vec<wgpu::CommandBuffer> {
            let resources: &TriangleRenderResources = resources.get().unwrap();
            resources.prepare(device, queue, self.angle);
            Vec::new()
        }

        fn paint<'a>(
            &self,
            _info: egui::PaintCallbackInfo,
            render_pass: &mut wgpu::RenderPass<'a>,
            resources: &'a egui_wgpu::CallbackResources,
        ) {
            let resources: &TriangleRenderResources = resources.get().unwrap();
            resources.paint(render_pass);
        }
    }
    pub struct TriangleRenderResources {
        pub pipeline: wgpu::RenderPipeline,
        pub bind_group: wgpu::BindGroup,
        pub uniform_buffer: wgpu::Buffer,
    }

    impl TriangleRenderResources {
        fn prepare(&self, _device: &wgpu::Device, queue: &wgpu::Queue, angle: f32) {
            // Update our uniform buffer with the angle from the UI
            queue.write_buffer(
                &self.uniform_buffer,
                0,
                bytemuck::cast_slice(&[angle, 0.0, 0.0, 0.0]),
            );
        }

        fn paint<'rp>(&'rp self, render_pass: &mut wgpu::RenderPass<'rp>) {
            // Draw our triangle!
            render_pass.set_pipeline(&self.pipeline);
            render_pass.set_bind_group(0, &self.bind_group, &[]);
            render_pass.draw(0..3, 0..1);
        }
    }
}
